import json
import dataclasses
import typing
import inflect

from collections import OrderedDict

from dataclasses_avroschema import schema_generator

p = inflect.engine()

BOOLEAN = "boolean"
NULL = "null"
INT = "int"
FLOAT = "float"
BYTES = "bytes"
STRING = "string"
ARRAY = "array"
ENUM = "enum"
MAP = "map"


PYTHON_TYPE_TO_AVRO = {
    bool: BOOLEAN,
    None: NULL,
    int: INT,
    float: FLOAT,
    bytes: BYTES,
    str: STRING,
    list: {"type": ARRAY},
    tuple: {"type": ENUM},
    dict: {"type": MAP},
}

# excluding tuple because is a container
PYTHON_INMUTABLE_TYPES = (str, int, bool, float, bytes)

PYTHON_PRIMITIVE_CONTAINERS = (list, tuple, dict)

PYTHON_PRIMITIVE_TYPES = PYTHON_INMUTABLE_TYPES + PYTHON_PRIMITIVE_CONTAINERS

PythonPrimitiveTypes = typing.Union[str, int, bool, float, list, tuple, dict]


@dataclasses.dataclass
class Field:
    name: str
    type: typing.Any  # store the python type (Field)
    default: typing.Any = dataclasses.MISSING
    default_factory: typing.Any = None

    # for avro array field
    items_type: typing.Any = None

    # for avro enum field
    symbols: typing.Any = None

    # for avro map field
    values_type: typing.Any = None

    # avro type storing
    avro_type: typing.Any = None

    def __post_init__(self):
        if isinstance(self.type, typing._GenericAlias) and not self.is_self_referenced(self.type):
            # Means that could be a list, tuple or dict
            origin = self.type.__origin__
            processor = self.get_processor(origin)
            processor()

            self.type = origin

    def get_processor(self, origin):
        """
        Get processor for a specific type.

        Supported: tuple, list, dict and typing.Type (custom types)
        """
        if origin is list:
            return self._process_list_type
        elif origin is dict:
            return self._process_dict_type
        elif origin is tuple:
            return self._process_tuple_type
        else:
            # we do not accept any other typing._GenericAlias like a set
            # we should raise an exception
            raise ValueError(
                f"Invalid Type for field {self.name}. Accepted types are list, tuple or dict")

    def _process_list_type(self):
        # because avro can have only one type, we take the first one
        items_type = self.type.__args__[0]

        if items_type in PYTHON_PRIMITIVE_TYPES:
            self.items_type = PYTHON_TYPE_TO_AVRO[items_type]
        elif self.is_self_referenced(items_type):
            # Checking for a self reference. Maybe is a typing.ForwardRef
            self.items_type = self._get_self_reference_type(items_type)
        else:
            # Is Avro Record Type
            self.items_type = schema_generator.SchemaGenerator(
                items_type).avro_schema_to_python()

    def _process_dict_type(self):
        """
        Process typing.Dict. Avro assumes that the key of a map is always a string,
        so we take the second argument to determine the value type
        """
        values_type = self.type.__args__[1]

        if values_type in PYTHON_PRIMITIVE_TYPES:
            self.values_type = PYTHON_TYPE_TO_AVRO[values_type]
        elif self.is_self_referenced(values_type):
            # Checking for a self reference. Maybe is a typing.ForwardRef
            self.values_type = self._get_self_reference_type(values_type)
        else:
            self.values_type = schema_generator.SchemaGenerator(
                values_type).avro_schema_to_python()

    def _process_tuple_type(self):
        self.symbols = list(self.default)

    @staticmethod
    def _get_self_reference_type(a_type):
        internal_type = a_type.__args__[0]

        return internal_type.__forward_arg__

    @staticmethod
    def is_self_referenced(a_type):
        return isinstance(a_type, typing._GenericAlias) \
            and a_type.__args__ and isinstance(a_type.__args__[0], typing.ForwardRef)

    @staticmethod
    def get_singular_name(name):
        singular = p.singular_noun(name)

        if singular:
            return singular
        return name

    def get_avro_type(self) -> PythonPrimitiveTypes:
        if self.is_self_referenced(self.type):
            return self._get_self_reference_type(self.type)

        avro_type = PYTHON_TYPE_TO_AVRO.get(self.type)

        if self.type in PYTHON_INMUTABLE_TYPES:
            if self.default is not dataclasses.MISSING and self.type is not tuple:
                if self.default is not None:
                    return [avro_type, NULL]
                # means that default value is None
                return [NULL, avro_type]

            return avro_type
        elif self.type in PYTHON_PRIMITIVE_CONTAINERS:
            if self.items_type:
                avro_type["items"] = self.items_type
            elif self.values_type:
                avro_type["values"] = self.values_type
            elif self.symbols:
                avro_type["symbols"] = self.symbols

            avro_type["name"] = self.get_singular_name(self.name)
            return avro_type
        else:
            # Assuming that is a Avro Record type.
            return schema_generator.SchemaGenerator(self.type).avro_schema_to_python()

    def get_default_value(self):
        if self.default is not dataclasses.MISSING:
            if self.type in PYTHON_INMUTABLE_TYPES:
                if self.default is None:
                    return NULL
                return self.default
            elif self.type is list:
                if self.default is None:
                    return []
            elif self.type is dict:
                if self.default is None:
                    return {}
        elif self.default_factory not in (dataclasses.MISSING, None):
            if self.type is list:
                # expeting a callable
                default = self.default_factory()
                assert isinstance(default, list), f"List is required as default for field {self.name}"

                return default
            elif self.type is dict:
                # expeting a callable
                default = self.default_factory()
                assert isinstance(default, dict), f"Dict is required as default for field {self.name}"

                return default

    def render(self) -> OrderedDict:
        """
        Render the fields base on the avro field

        At least will have name and type.

        returns:
            OrderedDict(
                ("name", "a name"),
                ("type", "a type")
            )

            The default key is optional.

            If self.type is:
                * list, the OrderedDict will contains the key items inside type
                * tuple, he OrderedDict will contains the key symbols inside type
                * dict, he OrderedDict will contains the key values inside type
        """
        template = OrderedDict([
            ("name", self.name),
            ("type", self.get_avro_type()),
        ])

        default = self.get_default_value()
        if default is not None:
            template["default"] = default

        return template

    def to_json(self) -> str:
        return json.dumps(self.render())

    def to_dict(self) -> dict:
        return json.loads(self.to_json())



import json
import dataclasses
import typing

from dataclasses_avroschema.schema_definition import AvroSchemaDefinition


class SchemaGenerator:

    def __init__(self, klass_or_instance, include_schema_doc: bool = True) -> None:
        self.dataclass = self.generate_dataclass(klass_or_instance)
        self.include_schema_doc = include_schema_doc
        self.schema_definition = None

    @staticmethod
    def generate_dataclass(klass_or_instance):
        if dataclasses.is_dataclass(klass_or_instance):
            return klass_or_instance
        return dataclasses.dataclass(klass_or_instance)

    def generate_schema(self, schema_type: str = "avro"):
        if self.schema_definition is not None:
            return self.schema_definition

        # let's live open the possibility to define different
        # schema definitions like json
        if schema_type == "avro":
            schema_definition = self._generate_avro_schema()
        else:
            raise ValueError("Invalid type. Expected avro schema type.")

        # cache the schema
        self.schema_definition = schema_definition

        # cache the schema
        self.schema_definition = schema_definition

        return self.schema_definition.render()

    def _generate_avro_schema(self) -> AvroSchemaDefinition:
        return AvroSchemaDefinition(
            "record",
            self.dataclass,
            include_schema_doc=self.include_schema_doc
        )

    def avro_schema(self) -> str:
        return json.dumps(self.generate_schema(schema_type="avro"))

    def avro_schema_to_python(self) -> typing.Dict[str, typing.Any]:
        return json.loads(self.avro_schema())

    @property
    def get_fields(self) -> typing.List["Field"]:
        if self.schema_definition is None:
            self.generate_schema()

        return self.schema_definition.fields



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" setup.py for dataclasses-avroschema."""

from setuptools import setup, find_packages

__version__ = "0.4.0"

with open("README.md") as readme_file:
    long_description = readme_file.read()

setup(
    name="dataclasses-avroschema",
    version=__version__,
    description="Generate Avro Schemas from a Python class",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Marcos Schroh",
    install_requires=["inflect==2.1.0"],
    author_email="schrohm@gmail.com",
    url="https://github.com/marcosschroh/dataclasses-avroschema",
    download_url="",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
    ],
    keywords=(
        """
        Python, Data Classes, Avro Schema, Avro, Apache, Data Streaming
        """
    ),
)



import dataclasses
import inspect
import typing
from collections import OrderedDict

from dataclasses_avroschema import fields


@dataclasses.dataclass
class BaseSchemaDefinition:
    """
    Minimal Schema definition
    """
    type: str
    klass_or_instance: dataclasses.dataclass

    def get_rendered_fields(self):
        raise NotImplementedError

    def render(self):
        raise NotImplementedError

    def get_schema_name(self):
        if inspect.isclass(self.klass_or_instance):
            return self.klass_or_instance.__name__
        return self.klass_or_instance.__class__.__name__

    def generate_documentation(self):
        doc = self.klass_or_instance.__doc__

        if doc is not None:
            return doc.replace("\n", "")


@dataclasses.dataclass
class AvroSchemaDefinition(BaseSchemaDefinition):
    aliases: typing.List[str] = None
    namespace: str = None
    fields: typing.List["fields.Field"] = None
    include_schema_doc: bool = True

    def __post_init__(self):
        self.generate_extra_avro_attributes()
        self.fields = self.parse_dataclasses_fields()

    def parse_dataclasses_fields(self) -> typing.List["fields.Field"]:
        return [
            fields.Field(
                dataclass_field.name,
                dataclass_field.type,
                dataclass_field.default,
                dataclass_field.default_factory
            )
            for dataclass_field in dataclasses.fields(self.klass_or_instance)
        ]

    def get_rendered_fields(self) -> typing.List["fields.Field"]:
        return [
            field.render() for field in self.fields
        ]

    def generate_extra_avro_attributes(self) -> None:
        """
        Look for the method in the dataclass.

        After calling the method extra_avro_attributes a dict is expected:
            typing.Dict[str, typing.Any]
        """
        extra_avro_attributes_fn = getattr(self.klass_or_instance, "extra_avro_attributes", None)

        if extra_avro_attributes_fn:
            extra_avro_attributes = extra_avro_attributes_fn()
            assert isinstance(extra_avro_attributes, dict), "Dict must be returned type in extra_avro_attributes method"

            aliases = extra_avro_attributes.get("aliases", self.aliases)
            namespace = extra_avro_attributes.get("namespace", self.namespace)

            self.aliases = aliases
            self.namespace = namespace

    def render(self):
        schema = OrderedDict([
            ("type", self.type),
            ("name", self.get_schema_name()),
            ("fields", self.get_rendered_fields())
        ])

        if self.include_schema_doc:
            doc = self.generate_documentation()
            if doc is not None:
                schema["doc"] = doc

        if self.namespace is not None:
            schema["namespace"] = self.namespace

        if self.aliases is not None:
            schema["aliases"] = self.aliases

        return schema
